#!/usr/bin/env python
#
# core: core nebulizer classes and functions

import sys
import os
import re
import fnmatch
import logging
import time
from bioblend import galaxy
from bioblend.galaxy.client import ConnectionError

logger = logging.getLogger(__name__)

class Credentials(object):
    """Class for managing credentials for Galaxy instances

    Credentials for different galaxy instances can be
    stored in a file under $HOME/.nebulizer

    Entries in the file consist of lines with three tab
    separated values:

    ALIAS   URL    API_KEY

    Blank lines or lines starting '#' are skipped.

    """

    def __init__(self,key_file=None):
        """
        Create a new Credentials instance

        Arguments:
          key_file (str): if supplied then should specify
            the path to the credentials file (defaults to
            $HOME/.nebulizer)

        """
        if key_file is None:
            key_file = os.path.join(os.path.expanduser("~"),
                                    '.nebulizer')
        self._key_file = os.path.abspath(key_file)

    def list_keys(self):
        """
        List aliases for API keys stored in credentials file

        Returns:
          List: list of aliases.

        """
        key_names = []
        if os.path.exists(self._key_file):
            with open(self._key_file,'r') as fp:
                for line in fp:
                    if line.startswith('#') or not line.strip():
                        continue
                    key_names.append(line.strip().split('\t')[0])
        return key_names

    def store_key(self,name,url,api_key):
        """
        Store a Galaxy API key

        Appends an entry to the key file.

        Arguments:
          name (str): alias to store the key against
          url (str): URL of the Galaxy instance
          api_key (str): API key for the Galaxy instance

        Returns:
          Boolean: True if key was updated, False on error.
        """
        if not api_key:
            logger.warning("Empty API key")
            return False
        with open(self._key_file,'a') as fp:
            fp.write("%s\t%s\t%s\n" % (name,url,api_key))
        return True

    def remove_key(self,name):
        """
        Remove a Galaxy API key

        Removes a key entry from the key file.

        Arguments:
          name (str) alias of the key to be removed

        Returns:
          Boolean: True if key was removed, False on error.
        """
        key_names = self.list_keys()
        if name not in key_names:
            logger.error("'%s': not found" % name)
            return False
        # Store the keys
        cached_keys = {
            name: list(self.fetch_key(name))
            for name in key_names
        }
        # Wipe the key file
        with open(self._key_file,'w') as fp:
            fp.write("#.nebulizer\n#Aliases\tGalaxy URL\tAPI key\n")
        # Store the cached keys again
        for alias in key_names:
            if name != alias:
                url,api_key = cached_keys[alias]
                self.store_key(alias,url,api_key)
        return True

    def update_key(self,name,new_url=None,new_api_key=None):
        """
        Update a Galaxy API key

        Updates the stored information in the key file.

        Arguments:
          name (str): alias of the key entry to be updated
          new_url (str): optional, new URL for the Galaxy
            instance
          new_api_key (str): optional, new API key for the
            Galaxy instance

        Returns:
          Boolean: True if key was updated, False on error.
        """
        try:
            url,api_key = self.fetch_key(name)
        except KeyError:
            logger.error("'%s': not found" % name)
            return False
        if new_url:
            url = new_url
        if new_api_key:
            api_key = new_api_key
        self.remove_key(name)
        self.store_key(name,url,api_key)
        return True

    def fetch_key(self,name):
        """
        Fetch credentials associated with a Galaxy instance

        Returns the credentials (i.e. Galaxy URL and API key)
        associated with the specified alias

        Raises a KeyError if no entry matching the alias is
        found.

        Arguments:
          name (str): alias of the key entry to fetch

        Returns:
          Tuple: consisting of (GALAXY_URL,API_KEY)
        """
        if os.path.exists(self._key_file):
            with open(self._key_file,'r') as fp:
                for line in fp:
                    if line.startswith('#') or not line.strip():
                        continue
                    alias,url,api_key = line.strip().split('\t')
                    if alias == name or url == name:
                        return (url,api_key)
        raise KeyError("'%s': not found" % name)

    def has_key(self,name):
        """
        Check if alias exists
        """
        try:
            self.fetch_key(name)
            return True
        except KeyError:
            return False

class Reporter(object):
    """
    Class for reporting column data

    Given multiple "lines" of column data (supplied
    as lists), pretty print the data in columns.

    Example usage:

    >>> output = Reporter()
    >>> output.append(['Some data',1.0,3])
    >>> output.append(['More stuff',21.9,19])
    >>> output.report()
    Some data   1.0   3
    More stuff  21.9  19
    """
    def __init__(self):
        """
        New Reporter instance
        """
        self._content = list()
        self._field_widths = list()
    def append(self,line):
        """
        Add a line of data

        Arguments:
          line (list): list of data items to
            append
        """
        self._content.append(line)
        for ix,item in enumerate(line):
            try:
                self._field_widths[ix] = max(self._field_widths[ix],
                                            len(str(item)))
            except IndexError:
                self._field_widths.append(len(str(item)))
    @property
    def nlines(self):
        """
        Number of lines stored
        """
        return len(self._content)
    def report(self,delimiter='  ',padding=True,prefix=None,
               rstrip=True):
        """
        Pretty-print the data

        Arguments:
          delimiter (str): delimiter to use (defaults
            to two space characters i.e. '  ')
          padding (bool): if True then line up columns
            of data by padding with spaces
          prefix (str): string to prepend to each line
          rstrip (bool): if True then strip all trailing
            whitespace from lines
        """
        output = []
        if padding:
            for line in self._content:
                # Apply padding to all but the last field
                out_line = ["%-*s" % (width,str(item))
                               for width,item
                               in zip(self._field_widths[:-1],
                                      line[:-1])]
                # Add the final field with no padding
                out_line.append(str(line[-1]))
                output.append(out_line)
        else:
            for line in self._content:
                output.append([str(item) for item in line])
        if not prefix:
            prefix = ''
        for line in output:
            out_line = "%s%s" % (prefix,delimiter.join(line))
            if rstrip:
                out_line = out_line.rstrip()
            print(out_line)

def get_galaxy_instance(galaxy_url,api_key=None,email=None,password=None,
                        verify_ssl=True):
    """
    Return Bioblend GalaxyInstance

    Attempts to connect to the specified Galaxy instance and
    verify that the connection is working by requesting the
    config for that instance.

    Arguments:
      galaxy_url (str): URL for the Galaxy instance to connect to
      api_key (str): API key to use when accessing Galaxy
      email (str): Galaxy e-mail address corresponding to the user
        (alternative to api_key; also need to supply a password)
      password (str): password of Galaxy account corresponding to
        email address (alternative to api_key)
      verify_ssl (bool): if True then turn off verification of SSL
        certificates for HTTPs connections

    Returns:
      GalaxyInstance: a bioblend GalaxyInstance for the connection,
        or None if the connection failed or couldn't be verified.

    """
    try:
        galaxy_url,stored_key = Credentials().fetch_key(galaxy_url)
    except KeyError as ex:
        logger.debug("Failed to find credentials for %s" %
                     galaxy_url)
        stored_key = None
    if api_key is None:
        api_key = stored_key
    logger.debug("Connecting to %s" % galaxy_url)
    if email is not None:
        gi = galaxy.GalaxyInstance(url=galaxy_url,email=email,
                                   password=password)
    else:
        gi = galaxy.GalaxyInstance(url=galaxy_url,key=api_key)
    gi.verify = verify_ssl
    if not get_galaxy_config(gi):
        return None
    user = get_current_user(gi)
    if user is not None:
        logger.debug("Connected as user %s" % user['email'])
    else:
        logger.debug("Unable to determine associated user")
    return gi

def get_galaxy_config(gi):
    """
    Requests configuration data for a Galaxy instance

    Arguments:
      gi (bioblend.galaxy.GalaxyInstance): Galaxy instance

    Returns:
      Dictionary: the configuration data for the Galaxy
        instance (will be empty if this couldn't be
        retrieved)
    """
    try:
        return galaxy.config.ConfigClient(gi).get_config()
    except ConnectionError as ex:
        print(ex)
        return {}

def get_current_user(gi):
    """
    Requests data on the user for an API connection

    Arguments:
      gi (bioblend.galaxy.GalaxyInstance): Galaxy instance

    Returns:
      Dictionary: the data on the user, or 'None' if the user
        couldn't be determined.
    """
    try:
        return galaxy.users.UserClient(gi).get_current_user()
    except ConnectionError:
        return None

def ping_galaxy_instance(gi):
    """
    Send a request to a Galaxy instance and report result

    Arguments:
      gi (bioblend.galaxy.GalaxyInstance): Galaxy instance

    Returns:
      Tuple: a tuple of the form (retcode,time). 'retcode'
        will be zero if the response from the server was okay,
        otherwise it is set to the status code of the failed
        response. 'time' is the time taken for the request to
        be sent and the response to be received, in seconds.

    """
    # Make a request
    try:
        start = time.time()
        galaxy.config.ConfigClient(gi).get_config()
        retcode = 0
    except ConnectionError as ex:
        retcode = ex.status_code
    end = time.time()
    return (retcode,end-start)

def prompt_for_confirmation(question,default=None):
    """
    Prompt the user to confirm an action

    Arguments:
      question: text to display next to the prompt
      default: default to use if user doesn't supply
        an explicit response (one of 'y','yes','n'
        or 'no')

    Returns:
      True if user confirms, False if not.

    """
    responses = {
        'yes': True,
        'ye' : True,
        'y'  : True,
        'no' : False,
        'n'  : False,
    }
    if default is None:
        prompt = " [y/n] "
    elif default.lower().startswith('y'):
        prompt = " [Y/n] "
    elif default.lower().startswith('n'):
        prompt = " [y/N] "
    else:
        raise Exception("Invalid default for prompt: %s" %
                        default)
    while True:
        sys.stdout.write(question + prompt)
        try:
            choice = input().lower()
        except NameError:
            choice = raw_input().lower()
        if default is not None and choice == '':
            return responses[default.lower()]
        elif choice in responses:
            return responses[choice]
        else:
            sys.stdout.write("Please respond with 'yes' or 'no' "
                             "(or 'y' or 'n').\n")

def turn_off_urllib3_warnings():
    """
    Turn off the warnings from urllib3

    Use this to suppress warnings (e.g. about unverified HTTPS
    certificates) that would otherwise be written out for each
    request in bioblend.

    """
    import requests
    requests.packages.urllib3.disable_warnings()



#!/usr/bin/env python

import unittest
from nebulizer.libraries import split_library_path
from nebulizer.libraries import normalise_folder_path

class TestSplitLibraryPathFunction(unittest.TestCase):
    """
    Tests for the 'split_library_path' function

    """
    def test_library_name_only(self):
        self.assertEqual(split_library_path('TestLibrary'),
                         ('TestLibrary','/'))
    def test_library_name_leading_slash(self):
        self.assertEqual(split_library_path('/TestLibrary'),
                         ('TestLibrary','/'))
    def test_library_name_trailing_slash(self):
        self.assertEqual(split_library_path('TestLibrary/'),
                         ('TestLibrary','/'))
    def test_library_name_with_spaces(self):
        self.assertEqual(split_library_path('Test Library 2'),
                         ('Test Library 2','/'))
    def test_library_and_folder(self):
        self.assertEqual(split_library_path('Test Library 2/Run 11'),
                         ('Test Library 2','/Run 11'))
    def test_library_and_folder_trailing_slash(self):
        self.assertEqual(split_library_path('Test Library 2/Run 11/'),
                         ('Test Library 2','/Run 11'))
    def test_library_and_multiple_folders(self):
        self.assertEqual(
            split_library_path('Test Library 2/Run 11/Fastqs'),
            ('Test Library 2','/Run 11/Fastqs'))
    def test_library_and_multiple_folders_trailing_slash(self):
        self.assertEqual(
            split_library_path('Test Library 2/Run 11/Fastqs/'),
            ('Test Library 2','/Run 11/Fastqs'))

class TestNormaliseFolderPath(unittest.TestCase):
    """
    Tests for the 'normalise_folder_path' function

    """
    def test_empty_folder_path(self):
        self.assertEqual(normalise_folder_path(''),'/')
    def test_already_normalised_folder_path(self):
        self.assertEqual(normalise_folder_path('/path/to/folder'),
                         '/path/to/folder')
    def test_folder_path_without_leading_slash(self):
        self.assertEqual(normalise_folder_path('path/to/folder'),
                         '/path/to/folder')
    def test_folder_path_with_trailing_slash(self):
        self.assertEqual(normalise_folder_path('/path/to/folder/'),
                         '/path/to/folder')
    def test_folder_path_slashes_reversed(self):
        self.assertEqual(normalise_folder_path('path/to/folder/'),
                         '/path/to/folder')
    def test_folder_path_remove_multiple_slashes(self):
        self.assertEqual(normalise_folder_path('/path//to/folder///'),
                         '/path/to/folder')




#!/usr/bin/env python
#
# search: functions for searching toolshed
import logging
import string
from fnmatch import fnmatch
from .core import get_galaxy_instance
from .core import Reporter
from .tools import normalise_toolshed_url
from .tools import get_repositories
from bioblend import toolshed
from bioblend import ConnectionError as BioblendConnectionError

# Logging
logger = logging.getLogger(__name__)

# Constants
SEARCH_PAGE_SIZE = 1000

# Functions

def search_toolshed(tool_shed,query_string,gi=None,
                    long_listing_format=False):
    """
    Search toolshed and print resulting matches

    Arguments:
      tool_shed (str): URL for tool shed to search
      query_string (str): text to use as query
      gi (bioblend.galaxy.GalaxyInstance): Galaxy instance
      long_listing_format (boolean): if True then use a
        long listing format when reporting items
    """
    # Get a toolshed instance
    tool_shed_url = normalise_toolshed_url(tool_shed)
    shed = toolshed.ToolShedInstance(tool_shed_url)
    print("Searching %s" % tool_shed_url)
    # Remove wildcards from start and end of query string
    shed_query_string = query_string.strip("*")
    # Query the toolshed
    repo_client = toolshed.repositories.ToolShedRepositoryClient(shed)
    try:
        search_result = repo_client.search_repositories(
            shed_query_string,
            page_size=SEARCH_PAGE_SIZE)
    except BioblendConnectionError as connection_error:
        # Handle error
        logger.warning("Error from Galaxy API: %s"
                       % connection_error)
        return connection_error.status_code
    # Filter on name
    hits = [r for r in search_result['hits'] if
            fnmatch(r["repository"]["name"].lower(),query_string)]
    # Deal with the results
    nhits = len(hits)
    if nhits == 0:
        print("No matching repositories found")
        return 0
    # Get additional details for each repo
    repositories = list()
    for hit in hits:
        # Get the repository details
        repo = hit['repository']
        name = repo['name']
        owner = repo['repo_owner_username']
        description = to_ascii(repo['description']).strip()
        # Get installable revisions
        installable_revisions = list()
        for revision in \
                repo_client.get_ordered_installable_revisions(name,owner):
            # Get details for each revision
            revision_info = \
                repo_client.get_repository_revision_install_info(
                    name,
                    owner,
                    revision)
            # Returns a 3 element list, only want details
            # from the last one
            # See https://bioblend.readthedocs.io/en/latest/api_docs/toolshed/all.html#bioblend.toolshed.repositories.ToolShedRepositoryClient.get_repository_revision_install_info
            revision_info = revision_info[2]
            version = revision_info[name][3]
            installable_revisions.append(dict(revision=revision,
                                              version=version,
                                              info=revision_info))
        # Sort the installable revisions on version number
        installable_revisions = sorted(installable_revisions,
                                       key=lambda r: int(r['version']),
                                       reverse=True)
        # Sort repo details
        repositories.append(dict(name=name,
                                 owner=owner,
                                 description=description,
                                 revisions=installable_revisions))
    # Get list of installed tool repositories
    if gi is not None:
        # Strip protocol from tool shed URL
        tool_shed = tool_shed_url
        for proc in ('http://','https://'):
            if tool_shed.startswith(proc):
                tool_shed = tool_shed[len(proc):]
        # Strip trailing slash
        tool_shed = tool_shed.rstrip('/')
        # Restrict repos to this tool shed
        installed_repos = [r for r in get_repositories(gi)
                           if r.tool_shed == tool_shed]
    else:
        installed_repos = []
    # Print the results
    print("")
    output = Reporter()
    for repository in repositories:
        # Get the repository details
        name = repository['name']
        owner = repository['owner']
        description = repository['description']
        # Iterate over revisions
        for revision in repository['revisions']:
            changeset = revision['revision']
            version = revision['version']
            # Look to see it's installed
            installed = bool([r for r in installed_repos
                              if (r.name == name and
                                  r.owner == owner and
                                  changeset in [rv.changeset_revision
                                                for rv in r.revisions()])])
            if installed:
                status = "*"
            else:
                status = " "
            # Print details
            if not long_listing_format:
                display_items = [owner,name,
                                 "%s:%s" % (version,changeset),
                                 status]
                output.append(display_items)
            else:
                output.append(("Name",name))
                output.append(("Owner",owner))
                output.append(("Revision","%s:%s" % (version,changeset)))
                output.append(("Description",description))
                if gi is not None:
                    if installed:
                        output.append(("Installed","yes"))
                    else:
                        output.append(("Installed","no"))
                output.append(("",))
    if not long_listing_format:
        output.report(prefix=" ")
    else:
        output.report(delimiter=": ")
    print("\n%d repositor%s found" % (nhits,('y' if nhits == 1 else 'ies')))
    # Finished
    return 0

def to_ascii(s,replace_with='?'):
    """
    Convert a string to ASCII

    Converts the supplied string to ASCII by replacing
    any non-ASCII characters with the specified
    character (defaults to '?').
    """
    try:
        return str(s)
    except UnicodeEncodeError:
        s1 = []
        for c in s:
            if c in string.printable:
                s1.append(c)
            else:
                s1.append(replace_with)
        return ''.join(s1)



#!/usr/bin/env python
#
# options: click definitions for shared options and arguments
import click

def install_tool_dependencies_option(default='yes'):
    return click.option('--install-tool-dependencies',
                        default=default,
                        type=click.Choice(['yes','no']),
                        help="install tool dependencies via the "
                        "toolshed, if any are defined (default "
                        "is '%s')" % default)

def install_repository_dependencies_option(default='yes'):
    return click.option('--install-repository-dependencies',
                        default=default,
                        type=click.Choice(['yes','no']),
                        help="install repository dependencies "
                        "via the toolshed, if any are defined "
                        "(default is '%s')" % default)

def install_resolver_dependencies_option(default='yes'):
    return click.option('--install-resolver-dependencies',
                        default=default,
                        type=click.Choice(['yes','no']),
                        help="install dependencies through a "
                        "resolver that supports installation "
                        "(e.g. conda) (default is '%s')" %
                        default)



# Current version of the library
__version__ = '0.6.0'

def get_version():
    """Returns a string with the current version of the library (e.g., "0.2.0")

    """
    return __version__

# Setup logging
import logging
logging.basicConfig()
logger = logging.getLogger("nebulizer")



#!/usr/bin/env python

import unittest
from nebulizer.users import User
from nebulizer.users import check_username_format
from nebulizer.users import get_username_from_login
from nebulizer.users import validate_password

class TestUser(unittest.TestCase):
    """
    Tests for the 'User' class

    """
    def test_load_user_data_minimal(self):
        # Data returned from galaxy.users.UserClient(gi).get_users()
        user_data = { u'username': u'bloggs',
                      u'model_class': u'User',
                      u'id': u'd6fbfd317568bb93',
                      u'email': u'joe.bloggs@galaxy.org' }
        user = User(user_data)
        self.assertEqual(user.username,'bloggs')
        self.assertEqual(user.email,'joe.bloggs@galaxy.org')
        self.assertEqual(user.id,'d6fbfd317568bb93')
    def test_load_user_data_full(self):
        # Data returned from galaxy.users.UserClient(gi).show_user()
        user_data = { u'username': u'bloggs',
                      u'quota_percent': 4,
                      u'total_disk_usage': 13181590307.0,
                      u'nice_total_disk_usage': u'12.3 GB',
                      u'id': u'd6fbfd317568bb93',
                      u'is_admin': True,
                      u'tags_used': [],
                      u'model_class': u'User',
                      u'email': u'joe.bloggs@galaxy.org' }
        user = User(user_data)
        self.assertEqual(user.username,'bloggs')
        self.assertEqual(user.email,'joe.bloggs@galaxy.org')
        self.assertEqual(user.id,'d6fbfd317568bb93')
        self.assertEqual(user.quota_percent,4)
        self.assertEqual(user.total_disk_usage,13181590307.0)
        self.assertTrue(user.is_admin)
        self.assertEqual(user.nice_total_disk_usage,'12.3 GB')

class TestCheckUsernameFormat(unittest.TestCase):
    def test_valid_username(self):
        self.assertTrue(check_username_format('peter-briggs'))
    def test_invalid_characters(self):
        self.assertFalse(check_username_format('J@ne!Doe'))
    def test_blank_username(self):
        self.assertFalse(check_username_format(''))

class TestGetUsernameFromLogin(unittest.TestCase):
    def test_get_username_from_login(self):
        self.assertEqual(get_username_from_login('joe.bloggs@galaxy.org'),
                         'joe-bloggs')

class TestValidatePassword(unittest.TestCase):
    def test_empty_password(self):
        self.assertFalse(validate_password(''))
    def test_invalid_password(self):
        self.assertFalse(validate_password('abc'))
    def test_valid_password(self):
        self.assertTrue(validate_password('p@55w0rd'))



#!/usr/bin/env python
#
# tools: functions for managing tools
import fnmatch
import time
import json
import logging
from bioblend import galaxy
from bioblend import toolshed
from bioblend.galaxy.client import ConnectionError
from bioblend import ConnectionError as BioblendConnectionError
from .core import prompt_for_confirmation
from .core import Reporter

# Logging
logger = logging.getLogger(__name__)

# Constants
TOOL_INSTALL_OK = 0
TOOL_INSTALL_FAIL = 1
TOOL_INSTALL_TIMEOUT = 2
TOOL_INSTALL_PENDING = 3
TOOL_UPDATE_OK = 0
TOOL_UPDATE_FAIL = 1
TOOL_UNINSTALL_OK = 0
TOOL_UNINSTALL_FAIL = 1

# Classes

class Tool(object):
    """
    Class wrapping extraction of tool data

    Provides an interface for accessing data about a tool
    in a Galaxy instance, which has been retrieved via a
    call to the Galaxy API using bioblend.

    """
    def __init__(self,tool_data):
        """
        Create a new Tool instance

        ``tool_data`` is a dictionary returned by a
        call to bioblend, for example:

        >>> for tool_data in galaxy.tools.ToolClient(gi).get_tools():
        >>>    print(Tool(tool_data).name)

        """
        self.name = tool_data['name']
        self.id = tool_data['id']
        self.version = tool_data['version']
        self.description = tool_data['description']
        self.panel_section = tool_data['panel_section_name']
        try:
            self.config_file = tool_data['config_file']
        except KeyError:
            self.config_file = None
        try:
            self.tool_shed_repository = tool_data['tool_shed_repository']
        except KeyError:
            self.tool_shed_repository = None

    @property
    def tool_repo(self):
        """
        Return the tool repository name

        The tool repository name is of the form

        TOOLSHED/OWNER/TOOL

        e.g. toolshed.g2.bx.psu.edu/devteam/tophat

        If the tool wasn't installed from a toolshed
        then an empty string is returned.

        """
        if self.tool_shed_repository:
            # Explicit repository information
            return '/'.join(
                [self.tool_shed_repository['tool_shed'],
                 self.tool_shed_repository['owner'],
                 self.tool_shed_repository['name']])
        # Older Galaxy: extract from the tool id
        ele = self.id.split('/')
        try:
            i = ele.index('repos')
            tool_shed = '/'.join(ele[:i])
            owner = ele[i+1]
            repo = ele[i+2]
            return '/'.join((tool_shed,owner,repo))
        except ValueError:
            return ''

    @property
    def tool_changeset(self):
        """
        Return the tool changeset revision

        This is a commit id of the form 'efc56ee1ade4'

        Returns None if a changeset revision can't be
        extracted

        """
        tool_repo = self.tool_repo
        if not tool_repo:
            return None
        if self.tool_shed_repository:
            # Explicit repository information
            return str(self.tool_shed_repository['changeset_revision'])
        # Older Galaxy: look for the config_file element - something
        # of the form:
        # .../toolshed.g2.bx.psu.edu/repos/devteam/picard/efc56ee1ade4/...
        try:
            ele = tool_repo.split('/')
            tool_shed = '/'.join(ele[:-2])
            owner = ele[-2]
            repo = ele[-1]
            search_string = "/repos/%s/%s/" % (owner,repo)
            i = self.config_file.index(search_string) + len(search_string)
            revision = self.config_file[i:].split('/')[0]
            return revision
        except (AttributeError,ValueError):
            return None

class RepositoryRevision(object):
    """
    Class wrapping extraction of toolshed repository version data

    Provides an interface for accessing data about the version
    data for a toolshed repository that has been installed into
    a Galaxy instance, and which has been retrieved via a call
    to the Galaxy API using bioblend.

    """
    def __init__(self,repo_data):
        """
        Create a new RepositoryRevision instance

        ``repo_data`` is a dictionary returned by a
        call to bioblend, for example:

        >>> for repo_data in \
        >>>   galaxy.toolshed.ToolShedClient(gi).get_repositories():
        >>>   print(RepositoryRevision(repo_data).revision_number)

        However the RepositoryRevision class is most
        likely instantiated implicitly by an instance
        of the Repository class.

        """
        # Version numbers
        self.revision_number = repo_data['ctx_rev']
        self.changeset_revision = repo_data['changeset_revision']
        self.installed_changeset_revision = repo_data['installed_changeset_revision']
        self.status = repo_data['status']
        self.error_message = repo_data['error_message']
        self.deleted = repo_data['deleted']
        # Repository revision status
        revision = repo_data['tool_shed_status']
        try:
            self.latest_revision = (revision['latest_installable_revision']
                                    == 'True')
        except (KeyError,TypeError):
            self.latest_revision = None
        try:
            self.revision_update = (revision['revision_update'] == 'True')
        except (KeyError,TypeError):
            self.revision_update = None
        try:
            self.revision_upgrade = (revision['revision_upgrade'] == 'True')
        except (KeyError,TypeError):
            self.revision_upgrade = None
        try:
            self.deprecated = (revision['repository_deprecated'] == 'True')
        except (KeyError,TypeError):
            self.deprecated = None
        # Indicates whether a newer revision is installed
        self._newer_revision_installed = None
        # Indicates whether a newer revision is available on toolshed
        self._tool_shed_has_newer_revision = None

    def newer_revision_installed(self,status=None):
        """
        Query or set flag indicating if newer version is installed

        If ``status`` is set then should be True
        (indicating that a newer revision of the tool
        has been installed in the Galaxy instance) or
        False (no newer instance has been installed).

        Returns the last value of ``status`` that was
        explicitly specified, or 'None' if no status
        has been set.
        """
        if status is not None:
            self._newer_revision_installed = status
        return self._newer_revision_installed

    def tool_shed_has_newer_revision(self,status=None):
        """
        Query or set flag indicating if newer version is available on toolshed

        If ``status`` is set then should be True
        (indicating that a newer version of the tool
        is available on the toolshed) or False (no
        newer version is available).

        Returns the last value of ``status`` that was
        explicitly specified, or 'None' if no status
        has been set.
        """
        if status is not None:
            self._tool_shed_has_newer_revision = status
        return self._tool_shed_has_newer_revision

    @property
    def status_indicator(self):
        """
        Return 'status_indicator' string for the revision

        The status indicator is a concise way to
        summarise the status of the revision:

        'D' = deprecated
        '^' = newer revision installed
        'u' = update available but not installed
        'U' = upgrade available but not installed
        '!' = newer version available on toolshed
        '*' = this is latest revision

        """
        status_indicator = ''
        if self.deprecated:
            status_indicator += 'D'
        elif self.newer_revision_installed():
            status_indicator += '^'
        elif self.tool_shed_has_newer_revision():
            status_indicator += '!'
        elif self.latest_revision:
            status_indicator += '*'
        elif self.revision_update:
            status_indicator += 'u'
        elif self.revision_upgrade:
            status_indicator += 'U'
        else:
            status_indicator = ' '
        return status_indicator

    @property
    def revision_id(self):
        """
        Return revision id

        The revision id is a combination of the
        numerical revision number (1,2...) and the
        revision changeset hash. For example:

        2:a60283899c6d

        """
        return ':'.join((self.revision_number,
                         self.changeset_revision))

class Repository(object):
    """
    Class wrapping extraction of toolshed repository data

    Provides an interface for accessing data about a
    toolshed repository that has been installed into
    a Galaxy instance, and which has been retrieved via
    a call to the Galaxy API using bioblend.

    """
    def __init__(self,repo_data):
        """
        Create a new Repository instance

        ``repo_data`` is a dictionary returned by a
        call to bioblend, for example:

        >>> for repo_data in \
        >>>   galaxy.toolshed.ToolShedClient(gi).get_repositories():
        >>>   print(Repository(repo_data).name)

        The Repository instance can store information
        on multiple revisions; additional revisions
        can be stored using the ``add_revision``
        method, and retrieved using the ``revisions``
        method.

        """
        self.name = repo_data['name']
        self.tool_shed = repo_data['tool_shed']
        self.owner = repo_data['owner']
        self._revisions = []
        self._tool_shed_revisions = None
        self.add_revision(repo_data)

    def add_revision(self,repo_data):
        """
        Add info on a specific revision

        ``repo_data`` is a dictionary returned by a
        call to bioblend.

        """
        self._revisions.append(RepositoryRevision(repo_data))
        # Sort into order (newest to oldest)
        self._revisions.sort(key=lambda r: int(r.revision_number),
                             reverse=True)
        # Find more recent installed version
        latest_installed_revision = None
        for i,r in enumerate(self._revisions):
            if not r.deleted:
                latest_installed_revision = \
                                self._revisions[i].changeset_revision
                break
        # Set status indicating whether newer version is
        # installed
        for r in self._revisions:
            r.newer_revision_installed(status=None)
        if latest_installed_revision is not None:
            self._revisions[i].newer_revision_installed(status=False)
            for r in self._revisions[i+1:]:
                r.newer_revision_installed(status=True)

    def update_tool_shed_revision_status(self):
        """
        Update the 'newer toolshed revision' status of the repository

        For each installed revision of the repository, set the
        ``tool_shed_has_newer_revision`` status to True if a
        newer revision is available on the toolshed, or False
        if not.
        """
        # Get the latest available revision from the toolshed
        tool_shed_revisions = self.tool_shed_revisions()
        if tool_shed_revisions is not None:
            latest_shed_revision = tool_shed_revisions[-1]
            if self._revisions[0].changeset_revision != latest_shed_revision:
                self._revisions[0].tool_shed_has_newer_revision(status=True)
            else:
                self._revisions[0].tool_shed_has_newer_revision(status=False)

    def tool_shed_revisions(self):
        """
        Fetch available repository revisions from toolshed

        Returns a list of the available installable revisions
        (aka changeset ids) from the toolshed, ordered oldest
        to newest.

        If the revision list can't be obtained (e.g. the tool
        shed is not available) then returns 'None', otherwise
        returns a list.
        """
        if self._tool_shed_revisions is None:
            # Fetch available revisions from the tool shed
            shed = toolshed.ToolShedInstance(url=self.tool_shed)
            try:
                self._tool_shed_revisions = \
                    shed.repositories.get_ordered_installable_revisions(
                        self.name,
                        self.owner)
            except BioblendConnectionError as connection_error:
                logger.critical("Unable to connect to toolshed '%s': %s" %
                                (self.tool_shed,connection_error.status_code))
        return self._tool_shed_revisions

    def revisions(self,include_deleted=False):
        """
        Return a list of installed revisions

        Arguments:
          include_deleted (bool): if True then include
            revisions marked as deleted (default is to
            omit deleted revisions)

        Returns:
          list: list of RepositoryRevision objects,
            in order of newest to oldest.

        """
        revisions = [x for x in self._revisions]
        if not include_deleted:
            revisions = [x for x in revisions if not x.deleted]
        return revisions

    @property
    def id(self):
        """
        Return an 'id' for the repo

        The id string is TOOL_SHED/OWNER/TOOL

        e.g. toolshed.g2.bx.psu.edu/devteam/tophat

        """
        return '/'.join((self.tool_shed,
                         self.owner,
                         self.name))

class ToolPanelSection(object):
    """
    Class wrapping extraction of tool panel sections

    Provides an interface for accessing data about a
    tool panel section in a Galaxy instance, and which
    has been retrieved via a call to the Galaxy API
    using bioblend.

    """
    def __init__(self,tool_panel_data):
        """
        Create a new ToolPanelSection instance

        ``tool_panel_data`` is a dictionary returned
        by a call to bioblend, for example:

        >>> galaxy.tools.ToolClient(gi)
        >>> for tool_data in galaxy.tools.ToolClient(gi).get_tools():
        >>>    print(ToolPanelSection(tool_data).name)

        """
        self.id = tool_panel_data['id']
        try:
            self.name = tool_panel_data['name']
        except KeyError:
            self.name = None
        self.model_class = tool_panel_data['model_class']
        self.elems = []
        try:
            for elem in tool_panel_data['elems']:
                self.elems.append(ToolPanelSection(elem))
        except KeyError:
            pass

    @property
    def is_toolsection(self):
        """
        Check if section is a tool panel section
        """
        return (self.model_class == "ToolSection")

    @property
    def is_tool(self):
        """
        Check if section is a tool
        """
        return (self.model_class == "Tool")

    @property
    def is_label(self):
        """
        Check if section is a tool panel label
        """
        return (self.model_class == "ToolSectionLabel")

class ToolPanel(object):
    """
    Class wrapping extraction of tool panel

    Provides an interface for accessing data about the
    tool panel in a Galaxy instance, and which
    has been retrieved via a call to the Galaxy API
    using bioblend.

    """
    def __init__(self,gi):
        """
        Create a new ToolPanel instance

        Arguments:
          gi (bioblend.galaxy.GalaxyInstance): Galaxy instance
        """
        self.sections = []
        tool_client = galaxy.tools.ToolClient(gi)
        for data in tool_client.get_tool_panel():
            self.sections.append(ToolPanelSection(data))

    def tool_index(self,tool,strict=False):
        """
        Return index of tool in tool panel

        Given a tool, return an integer 'index'
        corresponding to the position of the tool in
        the tool panel.

        By default if a tool at the specified
        version is not located then the index of the
        closest matching tool which matches the toolshed,
        owner and repository name but not the version
        will be returned instead.

        This is intended to deal with multiple versions
        of the same tool, where only one version (normally
        the most recent) appears in the tool panel.
        Other versions of the tool are accessible in Galaxy
        by selecting them from within the tool, so they
        could be considered to be 'hidden behind' the
        version that is explicitly installed.

        This behaviour can be disabled by setting the
        'strict' argument to True.

        Arguments:
          tool (Tool): Tool instance
          strict (boolean): if True then only return
            an index if there is a precise match to the
            tool (default is to try and return closest
            match if precise match cannot be found)

        Returns:
          Integer: position of the tool in the tool
            panel, or -1 if it can't be located.
        """
        # Convenience variables
        tool_id = tool.id
        tool_id_no_version = '/'.join(tool_id.split('/')[:-1])
        index_ = -1
        fallback_index = None
        # Traverse tool panel looking for a match
        for section in self.sections:
            if section.is_toolsection:
                for elem in section.elems:
                    index_ += 1
                    if elem.id == tool_id:
                        return index_
                    elif elem.id.startswith(tool_id_no_version) and \
                         section.name == tool.panel_section:
                        fallback_index = index_
            elif section.is_tool:
                index_ += 1
                if section.id == tool_id:
                    return index_
                elif section.id.startswith(tool_id_no_version):
                    fallback_index = index_
        # Not found, return nearest index instead
        if not strict and fallback_index is not None:
            return fallback_index
        # Nothing matching at all
        return -1

# Functions

def get_tools(gi):
    """
    Return list of tools in a Galaxy instance

    Arguments:
      gi (bioblend.galaxy.GalaxyInstance): Galaxy instance

    Returns:
      list: list of Tool objects.

    """
    tools = []
    tool_client = galaxy.tools.ToolClient(gi)
    for tool_data in tool_client.get_tools():
        tools.append(Tool(tool_data))
    return tools

def get_repositories(gi):
    """
    Return list of repositories installed in a Galaxy instance

    Arguments:
      gi (bioblend.galaxy.GalaxyInstance): Galaxy instance

    Returns:
      list: list of Repository objects.

    """
    repos = []
    shed_client = galaxy.toolshed.ToolShedClient(gi)
    for repo_data in shed_client.get_repositories():
        repo = Repository(repo_data)
        existing_repos = [x for x in repos if x.id == repo.id]
        if not existing_repos:
            # No entry for this repository
            repos.append(repo)
        elif len(existing_repos) == 1:
            # Single match, add this as a revision
            existing_repos[0].add_revision(repo_data)
        else:
            # Multiple matches - error
            raise Exception("Multiple instances for %s?" %
                            repo.id)
    return repos

def get_tool_panel_sections(gi):
    """
    Return list of tool panel sections in a Galaxy instance

    Arguments:
      gi (bioblend.galaxy.GalaxyInstance): Galaxy instance

    Returns:
      list: list of ToolPanelSection objects.

    """
    tool_panel_sections = []
    tool_client = galaxy.tools.ToolClient(gi)
    for data in tool_client.get_tool_panel():
        tool_panel_sections.append(ToolPanelSection(data))
    return tool_panel_sections

def get_revisions_from_toolshed(tool_shed,name,owner):
    """
    Fetch list of installable revisions for a tool from its toolshed

    Arguments:
      tool_shed (str): tool shed URL
      name (str): name of the tool repository
      owner (str): tool repository owner

    Returns:
      List: ordered list (oldest to newest) of the
        installable repositoy changesets on the tool
        shed.
    """
    shed = toolshed.ToolShedInstance(url=tool_shed)
    try:
        return shed.repositories.get_ordered_installable_revisions(name,
                                                                   owner)
    except BioblendConnectionError as connection_error:
        logger.critical("Unable to connect to toolshed '%s': %s" %
                        (tool_shed,connection_error.status_code))
        return []

def handle_repository_spec(repo_spec):
    """
    Process an arbitrary repository specification

    Given a repository specification as list of one of
    more components, returns the toolshed URL, repository
    owner, repository name and revision.

    The specification can be any of the forms:

    - ('https://toolshed.g2.bx.psu.edu/view/devteam/fastqc/e7b2202befea',)
    - (https://toolshed.g2.bx.psu.edu/view/devteam/fastqc',)
    - ('devteam/fastqc/e7b2202befea',)
    - ('devteam/fastqc',)
    - ('toolshed.g2.bx.psu.edu','devteam','fastqc','e7b2202befea')
    - ('toolshed.g2.bx.psu.edu','devteam','fastqc')
    - ('devteam','fastqc')
    - ('devteam','fastqc','e7b2202befea')
    - ('devteam','fastqc','3:e7b2202befea')

    Revision is set to None if not found; toolshed is set
    to the main Galaxy toolshed if not found.

    The leading protocol (https://, http://) is removed
    from the toolshed URL.

    Returns the tuple of (TOOLSHED,OWNER,REPOSITORY,REVISION)
    """
    repository = list(repo_spec)
    repo0 = repository[0].strip('/')
    if not (repo0.startswith('https://') or repo0.startswith('http://')):
        # No protocol specified
        if repo0.split('/')[0].count('.') > 0:
            # First element contains dots so must be a URL
            # Assume toolshed URL with no leading protocol
            repo0 = 'https://' + repo0
            if len(repository) > 1:
                # More than one element so assume that
                # first element is URL without trailing
                # tool definition
                # Check it ends with 'view' or 'repo'
                if not repo0.split('/')[-1] in ('view','repo'):
                    # Append 'view' element
                    repo0 += '/view'
        else:
            # No toolshed: assume main Galaxy toolshed
            repo0 = "https://toolshed.g2.bx.psu.edu/view/" + repo0
    repository[0] = repo0
    tool_url = '/'.join(repository)
    # Decompose the URL into toolshed, owner, repository
    # and changeset components
    toolshed = list()
    owner = None
    repository = None
    revision = None
    for ix,ele in enumerate(tool_url.split('/')):
        if ele not in ('view','repo'):
            toolshed.append(ele)
        else:
            toolshed = '/'.join(toolshed)
            try:
                owner = tool_url.split('/')[ix+1]
                repository = tool_url.split('/')[ix+2]
            except IndexError:
                # Invalid specification
                raise Exception("Invalid repository "
                                "specification: '%s'" %
                                repo_spec)
            try:
                revision = tool_url.split('/')[ix+3]
            except IndexError:
                revision = None
            break
    # Strip off the protocol
    for protocol in ('http://','https://'):
        if toolshed.startswith(protocol):
            toolshed = toolshed[len(protocol):]
            break
    # Handle revision of the form "version:changeset"
    # e.g. 3:f19e18ab01b1
    if revision and ':' in revision:
        revision = revision.split(':')[1]
    return (toolshed,owner,repository,revision)

def normalise_toolshed_url(tool_shed):
    """
    Return complete URL for a tool shed

    Arguments:
      tool_shed (str): partial or full URL for a
        toolshed server

    Returns:
      str: full URL for toolshed, including
        leading protocol.

    """
    if tool_shed.startswith('http://') or \
       tool_shed.startswith('https://'):
        return tool_shed
    return "https://%s" % tool_shed

def tool_install_status(gi,tool_shed,owner,name,revision=None):
    """
    Return the installation status of a tool repo

    If a revision is not specified then will return the
    status of the most recent installed version (if there
    is one).

    Arguments:
      gi (bioblend.galaxy.GalaxyInstance): Galaxy instance
      tool_shed (str): URL for the toolshed to install the
        tool from
      name (str): name of the tool repository
      owner (str): name of the tool repository owner
      revision (str): revision changeset specifying
        the tool repository version (optional)

    Returns:
      String: the tool repository installation status returned
        from Galaxy (or '?' if the status cannot be retrieved)

    """
    try:
        repos = get_repositories(gi)
    except ConnectionError as connection_error:
        logger.warning("Got connection error from Galaxy API: %s"
                       % connection_error.status_code)
        return "?"
    repos = [r for r in repos if (r.name == name and
                                  r.owner == owner and
                                  r.tool_shed == tool_shed)]
    if len(repos) != 1:
        logger.debug("Unable to fetch tool repository information")
        return "?"
    repo = repos[0]
    if revision:
        revisions = [v for v in repo.revisions()
                     if v.changeset_revision == revision]
    else:
        revisions = repo.revisions()
    if len(revisions) != 1:
        logger.debug("Unable to fetch tool repository revisions")
        return "?"
    rev = revisions[0]
    if rev.error_message:
        return rev.error_message
    return rev.status

def installed_repositories(gi,name=None,
                           tool_shed=None,
                           owner=None,
                           include_deleted=False,
                           only_updateable=False,
                           check_tool_shed=False):
    """
    Fetch a list of installed repository revisions

    Arguments:
      gi (bioblend.galaxy.GalaxyInstance): Galaxy instance
      name (str): optional, only list tool repositiories
        which match this string (can include wildcards)
      tool_shed (str): optional, only list tool
        repositories from toolsheds that match this string
        (can include wildcards)
      owner (str): optional, only list tool repositiories
        with owners who match this string (can include
        wildcards)
      list_tools (bool): if True then also list the tools
        provided by the repository
      include_deleted (bool): if True then also include
        repository revisions that are marked as deleted
        (default is to only show those which are not
        deleted)
      only_updateable (bool): if True then only report
        repositories that have uninstalled updates or
        upgrades available (default is to show all
        repositories and revisions)
      check_tool_shed (bool): if True then also check
        revisions against the tool shed, to determine if
        updates are available for each tool. NB this is
        an expensive operation to perform so is turned
        off by default

    Returns:
      List: a list of tuples consisting of three items
        (repository,revision,tools), where
        - 'repository' is a populated Repository instance
        - 'revision' is a populated RepositoryRevision
        - 'tools' is a list of Tool instances

    """
    # Get the list of installed repos
    installed_repos = []
    repos = get_repositories(gi)
    # Filter on name
    if name:
        name = name.lower()
        repos = [r for r in repos if fnmatch.fnmatch(r.name.lower(),name)]
    # Filter on toolshed
    if tool_shed:
        # Strip leading http(s)://
        for protocol in ('https://','http://'):
            if tool_shed.startswith(protocol):
                tool_shed = tool_shed[len(protocol):]
        repos = [r for r in repos if fnmatch.fnmatch(r.tool_shed,tool_shed)]
    # Filter on owner
    if owner:
        repos = [r for r in repos if fnmatch.fnmatch(r.owner,owner)]
    # Get list of tools
    tools = get_tools(gi)
    for repo in repos:
        # Also check against tool shed?
        if check_tool_shed:
            repo.update_tool_shed_revision_status()
        # Check each revision
        for revision in repo.revisions():
            # Exclude deleted revisions
            if not include_deleted and revision.deleted:
                continue
            # Exclude revisions that don't need updating
            if only_updateable and \
               (revision.newer_revision_installed() or
                (revision.latest_revision and
                 not revision.tool_shed_has_newer_revision())):
                continue
            # Fetch tools associated with this revision
            repo_tools = [t for t in tools
                          if (t.tool_repo == repo.id and
                              t.tool_changeset ==
                              revision.installed_changeset_revision)]
            # Append to the list
            installed_repos.append((repo,revision,repo_tools))
    # Finished
    return installed_repos

# Commands

def list_tools(gi,name=None,installed_only=False):
    """
    Print a list of the available tools

    Arguments:
      gi (bioblend.galaxy.GalaxyInstance): Galaxy instance
      name (str): optional, only list tools which match this
        string (can include wildcards)
      installed_only (bool): if True then only list those
        tools which are provided by toolshed repositories

    """
    tools = get_tools(gi)
    # Filter on name
    if name:
        name = name.lower()
        tools = [t for t in tools
                 if fnmatch.fnmatch(t.name.lower(),name)]
    # Filter on installed
    if installed_only:
        tools = [t for t in tools if t.tool_repo != '']
    # Sort into name order
    tools.sort(key=lambda x: x.name.lower())
    # Print info
    output = Reporter()
    for tool in tools:
        output.append((
            tool.name,tool.version,
            (tool.panel_section if tool.panel_section else ''),
            tool.tool_repo,
            (tool.tool_changeset if tool.tool_changeset else ''))
        )
    output.report()
    print("total %s" % len(tools))

def list_installed_repositories(gi,name=None,
                                tool_shed=None,
                                owner=None,
                                list_tools=False,
                                include_deleted=False,
                                only_updateable=False,
                                check_tool_shed=False,
                                tsv=False):
    """
    Print a list of the installed toolshed repositories

    Arguments:
      gi (bioblend.galaxy.GalaxyInstance): Galaxy instance
      name (str): optional, only list tool repositiories
        which match this string (can include wildcards)
      tool_shed (str): optional, only list tool
        repositories from toolsheds that match this string
        (can include wildcards)
      owner (str): optional, only list tool repositiories
        with owners who match this string (can include
        wildcards)
      list_tools (bool): if True then also list the tools
        provided by the repository
      include_deleted (bool): if True then also include
        repository revisions that are marked as deleted
        (default is to only show those which are not
        deleted)
      only_updateable (bool): if True then only report
        repositories that have uninstalled updates or
        upgrades available (default is to show all
        repositories and revisions)
      check_tool_shed (bool): if True then also check
        revisions against the tool shed, to determine if
        updates are available for each tool. NB this is
        an expensive operation to perform so is turned
        off by default
      tsv (bool): if True then output in a compact tab
        delimited format listing toolshed, owner,
        repository, changeset and tool panel section

    """
    # Get the list of installed repos
    repos = installed_repositories(gi,name=name,
                                   tool_shed=tool_shed,
                                   owner=owner,
                                   include_deleted=include_deleted,
                                   only_updateable=only_updateable,
                                   check_tool_shed=check_tool_shed)
    if tsv:
        # Output format for reinstallation of repositories
        tool_panel = ToolPanel(gi)
        # Sort into tool panel order
        repos = sorted(repos,
                       key=lambda r:
                       tool_panel.tool_index(r[2][0])
                       if r[2] else -1)
        # Filter out non-package, non-datamanager repositories
        # which can't be located in the tool panel
        repos = [r for r in repos if
                 (r[0].name.startswith("package_") or
                  r[0].name.startswith("data_manager_") or
                  (r[2] and tool_panel.tool_index(r[2][0]) > -1))]
        # Print details
        output = Reporter()
        for r in repos:
            repo,revision,tools = r
            if tools:
                tool_panel_section = tools[0].panel_section
            else:
                tool_panel_section = None
            output.append((repo.tool_shed,
                           repo.owner,
                           repo.name,
                           revision.changeset_revision,
                           (tool_panel_section
                            if tool_panel_section else '')))
        # Write out in TSV format
        output.report(delimiter='\t')
    else:
        # Denser more verbose format
        output = Reporter()
        nrevisions = 0
        for r in repos:
            # Print details
            repo,revision,tools = r
            output.append(('%s %s' % (revision.status_indicator,
                                      repo.name),
                           repo.tool_shed,
                           repo.owner,
                           revision.revision_id,
                           revision.status))
            nrevisions += 1
            # List tools associated with revision
            if list_tools:
                for tool in tools:
                    output.append(("-",
                                   tool.name,
                                   tool.version,
                                   tool.description))
        # Write to stdout
        output.report()
        print("total %s" % nrevisions)

def list_tool_panel(gi,name=None,list_tools=False):
    """
    Print a list of tool panel contents

    Arguments:
      gi (bioblend.galaxy.GalaxyInstance): Galaxy instance
      name (str): optional, only list sections which match
        this string (can include wildcards)
      list_tools (bool): if True then also print the
        tools under each tool panel section

    """
    # Get the list of tool panel sections
    tool_panel = ToolPanel(gi)
    # Filter on name
    if name:
        name = name.lower()
        sections = [s for s in tool_panel.sections
                    if fnmatch.fnmatch(s.name.lower(),name)]
    else:
        sections = tool_panel.sections
    # Get list of tools, if required
    if list_tools:
        tools = get_tools(gi)
    # Report
    output = Reporter()
    for section in sections:
        output.append(("'%s'" % section.name,
                       section.id))
        if list_tools:
            for tool in sorted([t for t in tools
                                if t.panel_section == section.name],
                               key=lambda t: tool_panel.tool_index(t)):
                output.append(("#%d" % tool_panel.tool_index(tool),
                               tool.name,
                               tool.version,
                               tool.description))
    output.report(rstrip=True)
    print("total %s" % len(sections))

def install_tool(gi,tool_shed,name,owner,revision=None,
                 tool_panel_section=None,
                 install_tool_dependencies=True,
                 install_repository_dependencies=True,
                 install_resolver_dependencies=True,
                 timeout=600,
                 poll_interval=10,no_wait=False):
    """
    Install a tool repository into a Galaxy instance

    Arguments:
      gi (bioblend.galaxy.GalaxyInstance): Galaxy instance
      tool_shed (str): URL for the toolshed to install the
        tool from
      name (str): name of the tool repository
      owner (str): name of the tool repository owner
      revision (str): optional revision changeset specifying
        the tool version to install; if not supplied then
        the latest installable revision will be used.
        NB revision can include the revision number e.g.
        "5:f8b7dc21b4ee"; only the trailing changeset id
        will be used.
      tool_panel_section (str): optional, name or id of
        the tool panel section to install the tools under; if
        the tool panel section doesn't already exist it will
        be created.
      install_tool_dependencies (bool): optional, if True
        then install tool dependencies from the toolshed
        if possible (default).
      install_repository_dependencies (bool): optional, if
        True then install repository dependencies from the
        toolshed if possible (default).
      install_resolver_dependencies (bool): optional, if
        True then install dependencies using a resolver
        which supports this (e.g. conda) (default).
      timeout (int): optional, sets the maximum time (in
        seconds) to wait for a tool to complete installing
        before giving up (default is 600s). Ignored if
        'no_wait' is True.
      poll_interval (int): optional, sets the time interval
        for polling Galaxy to check if a tool has completed
        installing (default is to check every 10s). Ignored
        if 'no_wait' is True.
      no_wait (boolean): optional, if True then don't wait
        for tool installation to complete (default is False
        i.e. do wait for tool to finish installing).

    """
    # Locate the repository on the toolshed
    print("Toolshed  :\t%s" % tool_shed)
    print("Repository:\t%s" % name)
    print("Owner     :\t%s" % owner)
    if revision is not None:
        # Normalise revision if necessary
        if ':' in revision:
            revision = revision.split(':')[1]
        print("Revision  :\t%s" % revision)
    else:
        print("Revision  :\t<not specified>")
    # Information on dependency installation
    print("Install tool dependencies from toolshed      : "
          "%s" % ('yes' if install_tool_dependencies else 'no'))
    print("Install repository dependencies from toolshed: "
          "%s" % ('yes' if install_repository_dependencies else 'no'))
    print("Install dependencies using resolver          : "
          "%s" % ('yes' if install_resolver_dependencies else 'no'))
    # Check if tool is already installed
    install_status = tool_install_status(gi,tool_shed,owner,name,
                                         revision)
    if install_status.startswith("Installed"):
        print("%s: already installed (status is \"%s\")" % (name,
                                                            install_status))
        return TOOL_INSTALL_OK
    # Get available revisions
    revisions = get_revisions_from_toolshed(tool_shed,name,owner)
    if not revisions:
        logger.critical("%s: no installable revisions found" % name)
        return TOOL_INSTALL_FAIL
    # Revisions are listed oldest to newest
    if revision is not None:
        # Check that specified revision can be installed
        if revision not in revisions:
            logger.critical("%s: requested revision is not installable"
                             % name)
            return TOOL_INSTALL_FAIL
    else:
        # Set revision to the most recent
        revision = revisions[-1]
        print("Installing newest revision (%s)" % revision)
    # Check if tool is already installed
    install_status = tool_install_status(gi,tool_shed,owner,name,
                                         revision)
    if install_status.startswith("Installed"):
        print("%s: already installed (status is \"%s\")" % (name,
                                                            install_status))
        return TOOL_INSTALL_OK
    # Look up tool panel section
    tool_panel_section_id = None
    new_tool_panel_section = None
    if tool_panel_section is None:
        print("Installing into top level tool panel (no section specified)")
    else:
        # Look for an existing section which matches the
        # supplied name or id
        for section in get_tool_panel_sections(gi):
            if tool_panel_section == section.id or \
               tool_panel_section == section.name:
                # Found existing tool panel section
                tool_panel_section_id = section.id
                print("Installing into existing tool panel section: "
                      "'%s' (id '%s')" % (section.name,
                                          tool_panel_section_id))
                break
        if not tool_panel_section_id:
            # No matching section exists, make a new one
            print("Installing into new tool panel section: '%s'" %
                  tool_panel_section)
            new_tool_panel_section = tool_panel_section
    # Get toolshed URL
    tool_shed_url = normalise_toolshed_url(tool_shed)
    print("Toolshed URL: %s" % tool_shed_url)
    # Attempt to install
    print("%s: requesting installation" % name)
    try:
        tool_shed_client = galaxy.toolshed.ToolShedClient(gi)
        tool_shed_client.install_repository_revision(
            tool_shed_url,name,owner,revision,
            install_tool_dependencies=install_tool_dependencies,
            install_repository_dependencies=install_repository_dependencies,
            install_resolver_dependencies=install_resolver_dependencies,
            tool_panel_section_id=tool_panel_section_id,
            new_tool_panel_section_label=new_tool_panel_section)
    except ConnectionError as connection_error:
        # Handle API error
        logger.warning("Got error from Galaxy API on attempted install "
                       "(ignored)")
        logger.warning("Status code: %s" % connection_error.status_code)
        logger.warning("Message    : \"%s\"" %
                       json.loads(connection_error.body)["err_msg"])
    except Exception as ex:
        # Handle general error
        logger.warning("Error while requesting tool installation "
                       "(ignored)")
        logger.warning("Exception: %s" % ex)
    # Monitor installation status
    if not no_wait:
        print("Galaxy connection closed: monitoring installation")
    ntries = 0
    prev_status_msg = None
    while (ntries*poll_interval) < timeout:
        install_status = tool_install_status(gi,tool_shed,owner,
                                             name,revision)
        if install_status.startswith("Installed"):
            print("%s: installed (status is \"%s\")" % (name,
                                                        install_status))
            return TOOL_INSTALL_OK
        elif install_status.startswith("Installing") or \
             install_status == "New" or \
             install_status == "Cloning" or \
             install_status == "Never installed" or \
             install_status == "?":
            if no_wait:
                # Don't wait for install to complete
                print("%s: still installing (status is \"%s\")" %
                      (name,install_status))
                print("Not waiting for install to complete")
                return TOOL_INSTALL_PENDING
            # Monitor the tool installation status
            ntries += 1
            status_msg = "%s: installing (status is \"%s\")" % (name,
                                                                install_status)
            if status_msg != prev_status_msg:
                print(status_msg)
                prev_status_msg = status_msg
            time.sleep(poll_interval)
        else:
            logger.critical("%s: failed (%s)" % (name,install_status))
            return TOOL_INSTALL_FAIL
    # Reaching here means timed out
    logger.critical("%s: timed out waiting for install" % name)
    return TOOL_INSTALL_TIMEOUT

def update_tool(gi,tool_shed,name,owner,
                install_tool_dependencies=True,
                install_repository_dependencies=True,
                install_resolver_dependencies=True,
                timeout=600,poll_interval=10,
                no_wait=False,check_tool_shed=False):
    """
    Update a tool repository in a Galaxy instance

    Arguments:
      gi (bioblend.galaxy.GalaxyInstance): Galaxy instance
      tool_shed (str): URL for the toolshed to install the
        tool from
      name (str): name of the tool repository
      owner (str): name of the tool repository owner
      install_tool_dependencies (bool): optional, if True
        then install tool dependencies from the toolshed
        if possible (default).
      install_repository_dependencies (bool): optional, if
        True then install repository dependencies from the
        toolshed if possible (default).
      install_resolver_dependencies (bool): optional, if
        True then install dependencies using a resolver
        which supports this (e.g. conda) (default).
      timeout (int): optional, sets the maximum time (in
        seconds) to wait for a tool to complete installing
        before giving up (default is 600s). Ignored if
        'no_wait' is True.
      poll_interval (int): optional, sets the time interval
        for polling Galaxy to check if a tool has completed
        installing (default is to check every 10s). Ignored
        if 'no_wait' is True.
      no_wait (boolean): optional, if True then don't wait
        for tool installation to complete (default is False
        i.e. do wait for tool to finish installing).
      check_tool_shed (bool): if True then also check
        revisions against the tool shed, to determine if
        updates are available for the tool (default is
        False i.e. do not check status against toolshed)
    """
    # Locate the existing installation
    update_repo = None
    for repo in get_repositories(gi):
        if repo.tool_shed == tool_shed and \
           repo.name == name and \
           repo.owner == owner:
            update_repo = repo
            break
    if update_repo is None:
        logger.critical("%s: unable to find repository to update" %
                        name)
        return TOOL_UPDATE_FAIL
    # Check there is at least one installed revision
    installed_revisions = [r for r in update_repo.revisions()
                           if not r.deleted]
    if not installed_revisions:
        logger.fatal("%s: no revisions currently installed" % name)
        return TOOL_UPDATE_FAIL
    # Update the toolshed status
    if check_tool_shed:
        update_repo.update_tool_shed_revision_status()
    # Find latest installable revision
    if not update_repo.tool_shed_revisions():
        logger.critical("%s: no installable revisions found" % name)
        return TOOL_UPDATE_FAIL
    revision = update_repo.tool_shed_revisions()[-1]
    # Check that there is an update available
    for r in update_repo.revisions():
        if not r.deleted and (r.latest_revision and
                              not r.tool_shed_has_newer_revision()):
            print("%s: version %s already the latest version" %
                  (name,r.revision_id))
            return TOOL_UPDATE_OK
    # Locate tool panel section for existing tools
    tool_panel_section = None
    for tool in get_tools(gi):
        if tool.tool_repo == update_repo.id:
            tool_panel_section = tool.panel_section
            break
    if tool_panel_section is None:
        logger.warning("%s: no tool panel section found" % name)
    #print("Installing update under %s" % tool_panel_section)
    return install_tool(
        gi,tool_shed,name,owner,revision,
        install_tool_dependencies=install_tool_dependencies,
        install_repository_dependencies=install_repository_dependencies,
        install_resolver_dependencies=install_resolver_dependencies,
        tool_panel_section=tool_panel_section,
        timeout=timeout,poll_interval=poll_interval,
        no_wait=no_wait)

def uninstall_tool(gi,tool_shed,name,owner,revision,
                   remove_from_disk=False,no_confirm=False):
    """
    Uninstall a tool repository from a Galaxy instance

    Arguments:
      gi (bioblend.galaxy.GalaxyInstance): Galaxy instance
      tool_shed (str): URL for the toolshed to install the
        tool from
      name (str): name of the tool repository
      owner (str): name of the tool repository owner
      revision (str): revision changeset to uninstall;
        if set to '*' then matches all installed
        revisions.
      remove_from_disk (bool): optional, if True then also
        remove the repository from disk (otherwise it is
        only deactivated).
      no_confirm : if True then don't prompt to confirm the
        uninstall operation.

    """
    # Locate existing tool to remove
    uninstall_repos = [r for r in get_repositories(gi)
                       if r.tool_shed == tool_shed
                       and r.name == name
                       and r.owner == owner]
    if not uninstall_repos:
        logger.fatal("%s/%s: no matching tool installed?" % (owner,name))
        return TOOL_UNINSTALL_FAIL
    elif len(uninstall_repos) > 1:
        logger.fatal("%s/%s: matches multiple installed tools?" %
                     (owner,name))
        return TOOL_UNINSTALL_FAIL
    else:
        uninstall_repo = uninstall_repos[0]
    # Identify revisions
    if revision is not None:
        remove_revisions = [r for r in uninstall_repo.revisions()
                            if revision == '*'
                            or revision == r.changeset_revision]
    else:
        remove_revisions = uninstall_repo.revisions()
        if len(remove_revisions) > 1:
            logger.fatal("%s/%s: no revision specified but multiple "
                         "revisions are installed" % (owner,name))
            return TOOL_UNINSTALL_FAIL
    if not remove_revisions:
        logger.fatal("%s/%s%s: no matching installed revision?" %
                     (owner,name,
                      '/%s' % revision if revision is not None else ''))
        return TOOL_UNINSTALL_FAIL
    # Get toolshed URL
    tool_shed_url = normalise_toolshed_url(tool_shed)
    print("Toolshed URL: %s" % tool_shed_url)
    # Report and confirm uninstall
    print("\nThe following tools will be uninstalled:\n")
    for r in remove_revisions:
        print("\t%s %s/%s %s" % (tool_shed,owner,name,r.revision_id))
    print("")
    if (not no_confirm) and \
       (not prompt_for_confirmation("Proceed?",default="n")):
        print("Uninstall cancelled")
        return TOOL_UNINSTALL_OK
    # Attempt to uninstall each revision
    uninstall_status = TOOL_UNINSTALL_OK
    for revision in remove_revisions:
        try:
            print("%s/%s: requesting uninstall" % (name,
                                                   revision.revision_id))
            tool_shed_client = galaxy.toolshed.ToolShedClient(gi)
            result = tool_shed_client.uninstall_repository_revision(
                name,owner,revision.changeset_revision,
                tool_shed_url,remove_from_disk=remove_from_disk)
            print("* %s" % result['message'])
        except ConnectionError as connection_error:
            # Handle API error
            logger.warning("Got error from Galaxy API on attempted uninstall "
                           "(ignored)")
            logger.warning(connection_error)
            logger.warning("Status code: %s" % connection_error.status_code)
            logger.warning("Message    : \"%s\"" %
                           json.loads(connection_error.body)["err_msg"])
        except Exception as ex:
            # Handle general error
            logger.warning("Error while requesting tool uninstall "
                           "(ignored)")
            logger.warning("Exception: %s" % ex)
            uninstall_status = TOOL_UNINSTALL_FAIL
    return uninstall_status



"""
Description

Setup script to install nebulizer: command line utilities for managing
users, tools and data libraries in Galaxy instances via the API

Copyright (C) University of Manchester 2015-2020 Peter Briggs

"""

readme = open('README.rst').read()

# Setup for installation etc
from setuptools import setup
import nebulizer
setup(
    name = "nebulizer",
    version = nebulizer.get_version(),
    description = "Manage users, tools and libraries in Galaxy",
    long_description = readme,
    url = 'https://github.com/pjbriggs/nebulizer',
    maintainer = 'Peter Briggs',
    maintainer_email = 'peter.briggs@manchester.ac.uk',
    packages = ['nebulizer',],
    entry_points = { 'console_scripts': [
        'nebulizer = nebulizer.cli:nebulizer',]
    },
    license = 'AFL',
    install_requires = ['bioblend>=0.13.0',
                        'mako',
                        'click<=6.7'],
    test_suite = 'nose.collector',
    tests_require = ['nose'],
    platforms="Posix; MacOS X; Windows",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: End Users/Desktop",
        "Intended Audience :: System Administrators",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Academic Free License (AFL)",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    include_package_data=True,
    zip_safe = False
)



#!/usr/bin/env python

import unittest
from nebulizer.tools import Tool
from nebulizer.tools import Repository
from nebulizer.tools import ToolPanelSection
from nebulizer.tools import handle_repository_spec
from nebulizer.tools import normalise_toolshed_url

class TestTool(unittest.TestCase):
    """
    Tests for the 'Tool' class

    """
    def test_load_tool_data(self):
        tool_data = { u'panel_section_name': u'NGS: Mapping',
                      u'description': u'Reports on methylation status of reads mapped by Bismark',
                      u'config_file': u'/galaxy/shed_tools/toolshed.g2.bx.psu.edu/repos/bgruening/bismark/0f8646f22b8d/bismark/bismark_bowtie_wrapper.xml',
                      u'name': u'Bismark Meth. Extractor',
                      u'panel_section_id': u'solexa_tools',
                      u'version': u'0.10.2',
                      u'link': u'/galaxy_dev/tool_runner?tool_id=toolshed.g2.bx.psu.edu%2Frepos%2Fbgruening%2Fbismark%2Fbismark_methylation_extractor%2F0.10.2',
                      u'min_width': -1,
                      u'model_class': u'Tool',
                      u'id': u'toolshed.g2.bx.psu.edu/repos/bgruening/bismark/bismark_methylation_extractor/0.10.2',
                      u'target': u'galaxy_main'}
        tool = Tool(tool_data)
        self.assertEqual(tool.name,'Bismark Meth. Extractor')
        self.assertEqual(tool.description,
                         'Reports on methylation status of reads mapped by '
                         'Bismark')
        self.assertEqual(tool.version,'0.10.2')
        self.assertEqual(tool.panel_section,'NGS: Mapping')
        self.assertEqual(tool.id,
                         'toolshed.g2.bx.psu.edu/repos/bgruening/bismark/'
                         'bismark_methylation_extractor/0.10.2')
        self.assertEqual(tool.tool_repo,
                         'toolshed.g2.bx.psu.edu/bgruening/bismark')
        self.assertEqual(tool.tool_changeset,'0f8646f22b8d')

    def test_load_tool_data_no_config_file(self):
        tool_data = { u'panel_section_name': u'NGS: Mapping',
                      u'description': u'Reports on methylation status of reads mapped by Bismark',
                      u'name': u'Bismark Meth. Extractor',
                      u'panel_section_id': u'solexa_tools',
                      u'version': u'0.10.2',
                      u'link': u'/galaxy_dev/tool_runner?tool_id=toolshed.g2.bx.psu.edu%2Frepos%2Fbgruening%2Fbismark%2Fbismark_methylation_extractor%2F0.10.2',
                      u'min_width': -1,
                      u'model_class': u'Tool',
                      u'id': u'toolshed.g2.bx.psu.edu/repos/bgruening/bismark/bismark_methylation_extractor/0.10.2',
                      u'target': u'galaxy_main'}
        tool = Tool(tool_data)
        self.assertEqual(tool.name,'Bismark Meth. Extractor')
        self.assertEqual(tool.description,
                         'Reports on methylation status of reads mapped by '
                         'Bismark')
        self.assertEqual(tool.version,'0.10.2')
        self.assertEqual(tool.panel_section,'NGS: Mapping')
        self.assertEqual(tool.id,
                         'toolshed.g2.bx.psu.edu/repos/bgruening/bismark/'
                         'bismark_methylation_extractor/0.10.2')
        self.assertEqual(tool.tool_repo,
                         'toolshed.g2.bx.psu.edu/bgruening/bismark')
        self.assertEqual(tool.tool_changeset,None)

    def test_load_tool_data_toolshed_has_prefix(self):
        tool_data = { u'panel_section_name': u'Local Toolshed',
                      u'config_file': u'/galaxy/shed_tools/192.168.60.164/toolshed/repos/pjbriggs/rnachipintegrator/2f0a1f1a5725/rnachipintegrator/rnachipintegrator_wrapper.xml',
                      u'description': u"Integrated analysis of 'gene' and 'peak' data",
                      u'panel_section_id': u'local_toolshed',
                      u'version': u'1.0.1.0',
                      u'link': u'/galaxy_dev/tool_runner?tool_id=192.168.60.164%2Ftoolshed%2Frepos%2Fpjbriggs%2Frnachipintegrator%2Frnachipintegrator_wrapper%2F1.0.1.0',
                      u'target': u'galaxy_main',
                      u'min_width': -1,
                      u'model_class': u'Tool',
                      u'id': u'192.168.60.164/toolshed/repos/pjbriggs/rnachipintegrator/rnachipintegrator_wrapper/1.0.1.0',
                      u'name': u'RnaChipIntegrator'}
        tool = Tool(tool_data)
        self.assertEqual(tool.name,'RnaChipIntegrator')
        self.assertEqual(tool.description,
                         "Integrated analysis of 'gene' and 'peak' data")
        self.assertEqual(tool.version,'1.0.1.0')
        self.assertEqual(tool.panel_section,'Local Toolshed')
        self.assertEqual(tool.id,
                         '192.168.60.164/toolshed/repos/pjbriggs/'
                         'rnachipintegrator/rnachipintegrator_wrapper/1.0.1.0')
        self.assertEqual(tool.tool_repo,
                         '192.168.60.164/toolshed/pjbriggs/'
                         'rnachipintegrator')
        self.assertEqual(tool.tool_changeset,'2f0a1f1a5725')

    def test_load_tool_data_with_tool_shed_repository_data(self):
        tool_data = { u'panel_section_name': u'NGS: SAMtools',
                      u'description': u'call variants',
                      u'name': u'MPileup',
                      u'labels': [],
                      u'edam_operations': [],
                      u'form_style': u'regular',
                      u'edam_topics': [],
                      u'panel_section_id': u'samtools',
                      u'version': u'2.1',
                      u'link': u'/tool_runner?tool_id=toolshed.g2.bx.psu.edu%2Frepos%2Fdevteam%2Fsamtools_mpileup%2Fsamtools_mpileup%2F2.1',
                      u'min_width': -1,
                      u'model_class': u'Tool',
                      u'id': u'toolshed.g2.bx.psu.edu/repos/devteam/samtools_mpileup/samtools_mpileup/2.1',
                      u'tool_shed_repository':
                      {u'owner': u'devteam',
                       u'changeset_revision':
                       u'820754ab8901',
                       u'name': u'samtools_mpileup',
                       u'tool_shed': u'toolshed.g2.bx.psu.edu'},
                      u'target': u'galaxy_main'}
        tool = Tool(tool_data)
        self.assertEqual(tool.name,'MPileup')
        self.assertEqual(tool.description,"call variants")
        self.assertEqual(tool.version,'2.1')
        self.assertEqual(tool.panel_section,'NGS: SAMtools')
        self.assertEqual(tool.id,
                         'toolshed.g2.bx.psu.edu/repos/devteam/'
                         'samtools_mpileup/samtools_mpileup/2.1')
        self.assertEqual(tool.tool_repo,
                         'toolshed.g2.bx.psu.edu/devteam/samtools_mpileup')
        self.assertEqual(tool.tool_changeset,'820754ab8901')

    def test_load_tool_data_not_from_toolshed(self):
        tool_data = { u'panel_section_name': u'Get Genomic Scores',
                      u'config_file': u'/galaxy/tools/filters/wiggle_to_simple.xml',
                      u'description': u'converter',
                      u'panel_section_id': u'scores',
                      u'version': u'1.0.0',
                      u'link': u'/galaxy_dev/tool_runner?tool_id=wiggle2simple1',
                      u'target': u'galaxy_main',
                      u'min_width': -1,
                      u'model_class': u'Tool',
                      u'id': u'wiggle2simple1',
                      u'name': u'Wiggle-to-Interval' }
        tool = Tool(tool_data)
        self.assertEqual(tool.name,'Wiggle-to-Interval')
        self.assertEqual(tool.description,
                         'converter')
        self.assertEqual(tool.version,'1.0.0')
        self.assertEqual(tool.panel_section,'Get Genomic Scores')
        self.assertEqual(tool.id,'wiggle2simple1')
        self.assertEqual(tool.tool_repo,'')
        self.assertEqual(tool.tool_changeset,None)

class TestRepository(unittest.TestCase):
    """
    """
    def test_load_repo_data(self):
        repo_data = { u'tool_shed_status':
                      { u'latest_installable_revision': u'True',
                        u'revision_update': u'False',
                        u'revision_upgrade': u'False',
                        u'repository_deprecated': u'False' },
                      u'status': u'Installed',
                      u'name': u'trimmomatic',
                      u'deleted': False,
                      u'ctx_rev': u'2',
                      u'error_message': u'',
                      u'installed_changeset_revision': u'a60283899c6d',
                      u'tool_shed': u'toolshed.g2.bx.psu.edu',
                      u'dist_to_shed': False,
                      u'url': u'/galaxy_dev/api/tool_shed_repositories/68b273cb7d2d6cff',
                      u'id': u'68b273cb7d2d6cff',
                      u'owner': u'pjbriggs',
                      u'uninstalled': False,
                      u'changeset_revision': u'a60283899c6d',
                      u'includes_datatypes': False }
        repo = Repository(repo_data)
        self.assertEqual(repo.name,'trimmomatic')
        self.assertEqual(repo.owner,'pjbriggs')
        self.assertEqual(repo.tool_shed,'toolshed.g2.bx.psu.edu')
        self.assertEqual(repo.id,
                         'toolshed.g2.bx.psu.edu/pjbriggs/trimmomatic')
        revisions = repo.revisions()
        self.assertEqual(len(revisions),1)
        self.assertEqual(revisions[0].revision_number,'2')
        self.assertEqual(revisions[0].changeset_revision,'a60283899c6d')
        self.assertEqual(revisions[0].installed_changeset_revision,
                         'a60283899c6d')
        self.assertEqual(revisions[0].status,'Installed')
        self.assertEqual(revisions[0].error_message,'')
        self.assertFalse(revisions[0].deleted)
        self.assertFalse(revisions[0].revision_update)
        self.assertFalse(revisions[0].revision_upgrade)
        self.assertTrue(revisions[0].latest_revision)
        self.assertEqual(revisions[0].revision_id,'2:a60283899c6d')

    def test_multiple_revisions(self):
        repo_data1 = { u'tool_shed_status':
                       { u'latest_installable_revision': u'True',
                         u'revision_update': u'False',
                         u'revision_upgrade': u'False',
                         u'repository_deprecated': u'False' },
                       u'status': u'Installed',
                       u'name': u'trimmomatic',
                       u'deleted': False,
                       u'ctx_rev': u'2',
                       u'error_message': u'',
                       u'installed_changeset_revision': u'a60283899c6d',
                       u'tool_shed': u'toolshed.g2.bx.psu.edu',
                       u'dist_to_shed': False,
                       u'url': u'/galaxy_dev/api/tool_shed_repositories/68b273cb7d2d6cff',
                       u'id': u'68b273cb7d2d6cff',
                       u'owner': u'pjbriggs',
                       u'uninstalled': False,
                       u'changeset_revision': u'a60283899c6d',
                       u'includes_datatypes': False }
        repo_data2 = { u'tool_shed_status':
                       {u'latest_installable_revision': u'False',
                        u'revision_update': u'False',
                        u'revision_upgrade': u'True',
                        u'repository_deprecated': u'False' },
                       u'status': u'Installed',
                       u'name': u'trimmomatic',
                       u'deleted': False,
                       u'ctx_rev': u'1',
                       u'error_message': u'',
                       u'installed_changeset_revision': u'3358c3d30143',
                       u'tool_shed': u'toolshed.g2.bx.psu.edu',
                       u'dist_to_shed': False,
                       u'url': u'/galaxy_dev/api/tool_shed_repositories/d6f760c242aa425c',
                       u'id': u'd6f760c242aa425c',
                       u'owner': u'pjbriggs',
                       u'uninstalled': False,
                       u'changeset_revision': u'2bd7cdbb6228',
                       u'includes_datatypes': False}
        repo = Repository(repo_data2)
        repo.add_revision(repo_data1)
        self.assertEqual(repo.name,'trimmomatic')
        self.assertEqual(repo.owner,'pjbriggs')
        self.assertEqual(repo.tool_shed,'toolshed.g2.bx.psu.edu')
        self.assertEqual(repo.id,
                         'toolshed.g2.bx.psu.edu/pjbriggs/trimmomatic')
        revisions = repo.revisions()
        # Most recent revision
        self.assertEqual(len(revisions),2)
        self.assertEqual(revisions[0].revision_number,'2')
        self.assertEqual(revisions[0].changeset_revision,'a60283899c6d')
        self.assertEqual(revisions[0].installed_changeset_revision,
                         'a60283899c6d')
        self.assertEqual(revisions[0].status,'Installed')
        self.assertEqual(revisions[0].error_message,'')
        self.assertFalse(revisions[0].deleted)
        self.assertFalse(revisions[0].revision_update)
        self.assertFalse(revisions[0].revision_upgrade)
        self.assertTrue(revisions[0].latest_revision)
        self.assertEqual(revisions[0].revision_id,'2:a60283899c6d')
        # Previous revision
        self.assertEqual(revisions[1].revision_number,'1')
        self.assertEqual(revisions[1].changeset_revision,'2bd7cdbb6228')
        self.assertEqual(revisions[1].installed_changeset_revision,
                         '3358c3d30143')
        self.assertEqual(revisions[1].status,'Installed')
        self.assertEqual(revisions[1].error_message,'')
        self.assertFalse(revisions[1].deleted)
        self.assertFalse(revisions[1].revision_update)
        self.assertTrue(revisions[1].revision_upgrade)
        self.assertFalse(revisions[1].latest_revision)
        self.assertEqual(revisions[1].revision_id,'1:2bd7cdbb6228')

class TestToolPanelSection(unittest.TestCase):
    """
    """
    def test_load_tool_panel_data_for_toolsection(self):
        tool_panel_data = { u'model_class': u'ToolSection',
                            u'version': u'',
                            u'elems':
                            [{u'panel_section_name': u'NGS: SAM Tools',
                              u'description': u'call variants',
                              u'name': u'MPileup',
                              u'panel_section_id': u'ngs:_sam_tools',
                              u'version': u'2.0',
                              u'link': u'/galaxy_dev/tool_runner?tool_id=toolshed.g2.bx.psu.edu%2Frepos%2Fdevteam%2Fsamtools_mpileup%2Fsamtools_mpileup%2F2.0',
                              u'min_width': -1,
                              u'model_class': u'Tool',
                              u'id': u'toolshed.g2.bx.psu.edu/repos/devteam/samtools_mpileup/samtools_mpileup/2.0',
                              u'target': u'galaxy_main'}],
                            u'name': u'NGS: SAM Tools',
                            u'id': u'ngs:_sam_tools' }
        section = ToolPanelSection(tool_panel_data)
        self.assertEqual(section.name,'NGS: SAM Tools')
        self.assertEqual(section.id,'ngs:_sam_tools')
        self.assertEqual(section.model_class,'ToolSection')
        self.assertTrue(section.is_toolsection)
        self.assertFalse(section.is_tool)
        self.assertFalse(section.is_label)
        self.assertEqual(len(section.elems),1)
        self.assertEqual(section.elems[0].name,'MPileup')
        self.assertEqual(section.elems[0].id,'toolshed.g2.bx.psu.edu/repos/devteam/samtools_mpileup/samtools_mpileup/2.0')
        self.assertEqual(section.elems[0].model_class,'Tool')
        self.assertFalse(section.elems[0].is_toolsection)
        self.assertTrue(section.elems[0].is_tool)
        self.assertFalse(section.elems[0].is_label)
        self.assertEqual(len(section.elems[0].elems),0)

    def test_load_tool_panel_data_for_tool(self):
        tool_panel_data = { u'panel_section_name': None,
                            u'config_file': u'/mnt/galaxy/shed_tools/toolshed.g2.bx.psu.edu/repos/devteam/fastqc/3fdc1a74d866/fastqc/rgFastQC.xml',
                            u'description': u'Read Quality reports',
                            u'labels': [],
                            u'name': u'FastQC',
                            u'panel_section_id': None,
                            u'version': u'0.65',
                            u'link': u'/tool_runner?tool_id=toolshed.g2.bx.psu.edu%2Frepos%2Fdevteam%2Ffastqc%2Ffastqc%2F0.65',
                            u'min_width': -1,
                            u'model_class': u'Tool',
                            u'id': u'toolshed.g2.bx.psu.edu/repos/devteam/fastqc/fastqc/0.65',
                            u'target': u'galaxy_main' }
        section = ToolPanelSection(tool_panel_data)
        self.assertEqual(section.name,'FastQC')
        self.assertEqual(section.id,'toolshed.g2.bx.psu.edu/repos/devteam/fastqc/fastqc/0.65')
        self.assertEqual(section.model_class,'Tool')
        self.assertFalse(section.is_toolsection)
        self.assertTrue(section.is_tool)
        self.assertFalse(section.is_label)
        self.assertEqual(len(section.elems),0)

    def test_load_tool_panel_data_for_label(self):
        tool_panel_data = { u'model_class': u'ToolSectionLabel',
                            u'version': u'',
                            u'id': u'deprecated',
                            u'text': u'DEPRECATED' }
        section = ToolPanelSection(tool_panel_data)
        self.assertEqual(section.name,None)
        self.assertEqual(section.id,'deprecated')
        self.assertEqual(section.model_class,'ToolSectionLabel')
        self.assertFalse(section.is_toolsection)
        self.assertFalse(section.is_tool)
        self.assertTrue(section.is_label)
        self.assertEqual(len(section.elems),0)

class TestHandleRepositorySpec(unittest.TestCase):
    """
    Tests for the 'handle_repository_spec' function
    """
    def test_handle_repository_spec_full_url(self):
        self.assertEqual(
            handle_repository_spec(("https://toolshed.g2.bx.psu.edu/view/devteam/fastqc/e7b2202befea",)),
            ("toolshed.g2.bx.psu.edu",
             "devteam",
             "fastqc",
             "e7b2202befea"
            ))
        self.assertEqual(
            handle_repository_spec(("https://toolshed.g2.bx.psu.edu/view/devteam/fastqc",)),
            ("toolshed.g2.bx.psu.edu",
             "devteam",
             "fastqc",
             None
            ))
    def test_handle_repository_spec_full_url(self):
        self.assertEqual(
            handle_repository_spec(("https://local.org/toolshed/view/devteam/fastqc/e7b2202befea",)),
            ("local.org/toolshed",
             "devteam",
             "fastqc",
             "e7b2202befea"
            ))
    def test_handle_repository_spec_owner_slash_tool(self):
        self.assertEqual(
            handle_repository_spec(("devteam/fastqc",)),
            ("toolshed.g2.bx.psu.edu",
             "devteam",
             "fastqc",
             None
            ))
    def test_handle_repository_spec_owner_space_tool(self):
        self.assertEqual(
            handle_repository_spec(("devteam","fastqc",)),
            ("toolshed.g2.bx.psu.edu",
             "devteam",
             "fastqc",
             None
            ))
    def test_handle_repository_spec_space_separated(self):
        self.assertEqual(
            handle_repository_spec(("toolshed.g2.bx.psu.edu",
                                    "devteam","fastqc",)),
            ("toolshed.g2.bx.psu.edu",
             "devteam",
             "fastqc",
             None
            ))
        self.assertEqual(
            handle_repository_spec(("toolshed.g2.bx.psu.edu",
                                    "devteam","fastqc",
                                    "e7b2202befea",)),
            ("toolshed.g2.bx.psu.edu",
             "devteam",
             "fastqc",
             "e7b2202befea"
            ))
        self.assertEqual(
            handle_repository_spec(("toolshed.g2.bx.psu.edu",
                                    "devteam","fastqc",
                                    "3:e7b2202befea",)),
            ("toolshed.g2.bx.psu.edu",
             "devteam",
             "fastqc",
             "e7b2202befea"
            ))
    def test_handle_repository_spec_invalid_spec_raises_exception(self):
        self.assertRaises(
            Exception,
            handle_repository_spec,
            ("fastqc",))

class TestNormaliseToolshedUrl(unittest.TestCase):
    """
    Tests for the 'normalise_toolshed_url' function

    """
    def test_full_https_url(self):
        self.assertEqual(
            normalise_toolshed_url('https://toolshed.g2.bx.psu.edu'),
            'https://toolshed.g2.bx.psu.edu')
    def test_full_http_url(self):
        self.assertEqual(
            normalise_toolshed_url('http://127.0.0.1:9009'),
            'http://127.0.0.1:9009')
    def test_no_protocol(self):
        self.assertEqual(
            normalise_toolshed_url('127.0.0.1:9009'),
            'https://127.0.0.1:9009')
